package android_uart_api;

import anywheresoftware.b4a.BA;
//import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.objects.streams.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.zip.Adler32;

  
@BA.Hide
public class AsyncStreams {  
	
	static final int STREAM_PREFIX = -2;
	static final byte[] CLOSE_PILL = new byte[0];
	private String eventName;
	private BA ba;
	private AsyncStreams.AIN ain;
	private Thread tin;
	private AsyncStreams.AOUT aout;
	private Thread tout;
	public String StreamFolder;
	volatile long streamReceived;
	volatile long streamTotal;
    

	public void setAsyncStreams(BA ba, String EventName, InputStream In, OutputStream Out, boolean BigEndian, boolean Prefix) throws IOException {
		this.shared(ba, In, Out, EventName, BigEndian, Prefix);
	}

  
	/*
	public void Initialize(BA ba, InputStream In, OutputStream Out, String EventName) throws IOException {
		this.shared(ba, In, Out, EventName, false, false);
	}

    
	public void InitializePrefix(BA ba, InputStream In, boolean BigEndian, OutputStream Out, String EventName) throws IOException {
		if (File.getExternalWritable()) {
			this.StreamFolder = File.getDirDefaultExternal();
    	} 
    	else {
    		this.StreamFolder = File.getDirInternalCache();
    	}

    	this.shared(ba, In, Out, EventName, BigEndian, true);      
	}
	*/

    
	public long getStreamTotal() {
    	return this.streamTotal;
    }

    
    public long getStreamReceived() {
    	return this.streamReceived;
    }

      
    private void shared(BA ba, InputStream In, OutputStream Out, String EventName, boolean BigEndian, boolean Prefix) throws IOException {
    	if (this.IsInitialized()) {
    		this.Close();
    	}

    	this.ba = ba;
    	this.eventName = EventName.toLowerCase(BA.cul);
    	
    	if (In != null) {
    		this.ain = new AsyncStreams.AIN(In, BigEndian, Prefix);
    		this.tin = new Thread(this.ain);
    		this.tin.setDaemon(true);
    		this.tin.start();
    	}

    	if (Out != null) {
    		this.aout = new AsyncStreams.AOUT(Out, BigEndian, Prefix);
    		this.tout = new Thread(this.aout);
    		this.tout.setDaemon(true);
    		this.tout.start();
    	}
    }

    
    public boolean IsInitialized() {
    	return this.ain != null || this.aout != null;
    }

    
    public boolean Write(byte[] Buffer) {
    	return this.Write2(Buffer, 0, Buffer.length);
    }

    
    public boolean Write2(byte[] Buffer, int Start, int Length) {
    	AsyncStreams.AOUT a = this.aout;
    	return a == null ? false : a.put(Buffer, Start, Length);
    }

    
    public boolean SendAllAndClose() {
    	return this.Write2(CLOSE_PILL, 0, 0);
    }

    
    public boolean WriteStream(InputStream In, Long Size) {
    	AsyncStreams.AOUT a = this.aout;
    	return a == null ? false : a.put(In, Size);
    }

    
    public int getOutputQueueSize() {
    	return this.aout == null ? 0 : this.aout.queue.size();
    }

   
    public synchronized void Close() throws IOException {
    	if (this.tin != null && this.ain != null) {
    		this.ain.close();
    		if (Thread.currentThread() != this.tin) {
    			this.tin.interrupt();
    		}
    	}

    	if (this.tout != null && this.aout != null) {
    		this.aout.close();
    		if (Thread.currentThread() != this.tout) {
    			this.tout.interrupt();
    		}
    	}

    	this.ain = null;
    	this.aout = null;
    }

   
    private class AIN implements Runnable {
    	private final InputStream in;
    	private byte[] buffer = new byte[8192];
    	private final byte[] prefixBuffer = new byte[4];
    	private final boolean prefix;
    	private volatile boolean working = true;
    	private String ev;
    	private ByteBuffer bb;

    	public AIN(InputStream in, boolean bigEndian, boolean prefix) {
    		this.ev = AsyncStreams.this.eventName + "_newdata";
    		this.in = in;
    		this.prefix = prefix;
    		if (prefix) {
    			this.bb = ByteBuffer.wrap(new byte[8]);
    			this.bb.order(bigEndian ? ByteOrder.BIG_ENDIAN : ByteOrder.LITTLE_ENDIAN);
    		}
    	}

    	public void run() {
    		try {
    			while(this.working) {
    				byte[] data;
    				int msgLength;
    				if (!this.prefix) {
    					msgLength = this.in.read(this.buffer);
    					
    					if (msgLength == 0) {
    						continue;
    					}
    					
    					if (msgLength < 0) {
    						this.closeUnexpected();
    						break;
    					}

    					if (!this.working) {
    						break;
    					}

    					data = new byte[msgLength];
    					System.arraycopy(this.buffer, 0, data, 0, msgLength);
    				} 
    				else {
    					if (!this.readNumberOfBytes(this.in, this.prefixBuffer, 4) || !this.working) {
    						break;
    					}

    					this.bb.clear();
    					this.bb.put(this.prefixBuffer, 0, 4);
    					msgLength = this.bb.getInt(0);
    					
    					if (msgLength > 100000000) {
    						throw new RuntimeException("Message size too large. Prefix mode can only work if both sides of the connection follow the 'prefix' protocol.");
    					}

    					if (msgLength == -2)
    					{
    						if (!this.readNumberOfBytes(this.in, this.buffer, 8))
    						{
    							break;
    						}

    						this.bb.clear();
    						this.bb.put(this.buffer, 0, 8);
    						AsyncStreams.this.streamTotal = this.bb.getLong(0);
    						AsyncStreams.this.streamReceived = 0L;

    						int i;
    						for(i = 1; File.Exists(AsyncStreams.this.StreamFolder, String.valueOf(i)); ++i) {
    							;
    						}

    						OutputStream out = (OutputStream)File.OpenOutput(AsyncStreams.this.StreamFolder, String.valueOf(i), false).getObject();

    						try {
    							Adler32 adler;
    							long crc;
    							int len;
    							for(adler = new Adler32(); AsyncStreams.this.streamReceived < AsyncStreams.this.streamTotal; AsyncStreams.this.streamReceived += (long)len) {
    								crc = AsyncStreams.this.streamTotal - AsyncStreams.this.streamReceived;
    								len = (int)(crc > (long)this.buffer.length ? (long)this.buffer.length : crc);
    								
    								if (!this.readNumberOfBytes(this.in, this.buffer, len)) {
    									break;
    								}

    								adler.update(this.buffer, 0, len);
    								out.write(this.buffer, 0, len);
    							}

    							if (!this.readNumberOfBytes(this.in, this.buffer, 8)) {
    								break;
    							}

    							this.bb.clear();
    							this.bb.put(this.buffer, 0, 8);
    							crc = this.bb.getLong(0);
    							
    							if (crc != adler.getValue()) {
    								throw new Exception("CRC value does not match.");
    							}
    						} 
    						finally {
    							out.close();
    						}

    						AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, AsyncStreams.this.eventName + "_newstream", true, new Object[]{AsyncStreams.this.StreamFolder, String.valueOf(i)});
    						continue;
    					}

    					if (msgLength > this.buffer.length) {
    						this.buffer = new byte[msgLength];
    					}

    					if (!this.readNumberOfBytes(this.in, this.buffer, msgLength)) {
    						break;
    					}

                  data = new byte[msgLength];
                  System.arraycopy(this.buffer, 0, data, 0, data.length);
               }

               AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, this.ev, true, new Object[]{data});
            }
         } catch (Exception ex) {
            if (this.working) {
               ex.printStackTrace();
               AsyncStreams.this.ba.setLastException(ex);
               AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, AsyncStreams.this.eventName + "_error", false, (Object[])null);
            }
         }

      }

      private boolean readNumberOfBytes(InputStream in, byte[] buffer, int len) throws IOException {
         int c;
         for(int count = 0; count < len; count += c) {
            c = in.read(buffer, count, len - count);
            if (c == -1) {
               this.closeUnexpected();
               return false;
            }
         }

         return true;
      }

      private void closeUnexpected() throws IOException {
         AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, AsyncStreams.this.eventName + "_terminated", false, (Object[])null);
         AsyncStreams.this.Close();
      }

      public void close() {
         this.working = false;

         try {
            this.in.close();
         } catch (IOException ex) {
            ex.printStackTrace();
         }
      }
   }


    private class AOUT implements Runnable {
    	private final OutputStream out;
    	private volatile boolean working = true;
    	private final ArrayBlockingQueue<Object> queue = new ArrayBlockingQueue<Object>(100);
    	private final boolean prefix;
    	private final ByteBuffer bb;
    	private byte[] streamBuffer;

    	public AOUT(OutputStream out, boolean bigEndian, boolean prefix) {
    		this.out = out;
    		this.prefix = prefix;
    		if (prefix) {
    			this.bb = ByteBuffer.wrap(new byte[8]);
    			this.bb.order(bigEndian ? ByteOrder.BIG_ENDIAN : ByteOrder.LITTLE_ENDIAN);
    		}
    		else {
    			this.bb = null;
    		}
    	}
   	   	
    	public void run() {
    		while(this.working)	{
    			try	{
    				Object b = this.queue.take(); 
    				if (b instanceof byte[]) {
    					if (b == AsyncStreams.CLOSE_PILL) {
    						AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, AsyncStreams.this.eventName + "_terminated", false, (Object[])null);
    						AsyncStreams.this.Close();
    						return;
    					}
    					this.out.write((byte[])b);
    				} 
    				else {
    					AsyncStreams.StreamAndSize st = (AsyncStreams.StreamAndSize)b;
    					try {
    						Adler32 adler = new Adler32();
    						synchronized(this.bb) {
    							this.bb.putInt(0, -2);
    							this.out.write(this.bb.array(), 0, 4);
    							this.bb.putLong(0, st.size);
    							this.out.write(this.bb.array(), 0, 8);
    						}
    						
    						if (this.streamBuffer == null) {
    							this.streamBuffer = new byte[8192];
    						}

    						int len;
    						while((len = st.in.read(this.streamBuffer)) > 0) {
    							this.out.write(this.streamBuffer, 0, len);
    							adler.update(this.streamBuffer, 0, len);
    						}
    						synchronized(this.bb) {
    							this.bb.putLong(0, adler.getValue());
    							this.out.write(this.bb.array(), 0, 8);
    						}
    					} 
    					finally {
    						try {
    							st.in.close();
    						}
    						catch (Exception ex) {
    							ex.printStackTrace();
    						}
    					}
    				}
    			} 
    			catch (Exception ex) {
    				if (this.working) {
    					ex.printStackTrace();
    					AsyncStreams.this.ba.setLastException(ex);
    					AsyncStreams.this.ba.raiseEventFromDifferentThread(AsyncStreams.this, (Object)null, 0, AsyncStreams.this.eventName + "_error", false, (Object[])null);
    				}
    			}
    		}
    	}

    	public boolean put(InputStream in, long size) {
    		if (!this.prefix) {
    			throw new RuntimeException("WriteStream is only supported in prefix mode.");
    		} 
    		else {
    			try {
    				AsyncStreams.StreamAndSize st = new AsyncStreams.StreamAndSize();
    				st.in = in;
    				st.size = size;
    				return this.queue.offer(st, 100L, TimeUnit.MILLISECONDS);
    			}
    			catch (InterruptedException ex) {
    				throw new RuntimeException(ex);
    			}
    		}
    	}
   
      	public boolean put(byte[] buffer, int start, int len) {
      		byte[] b;
      		if (buffer == AsyncStreams.CLOSE_PILL) {
      			b = AsyncStreams.CLOSE_PILL;
      		} 
      		else if (!this.prefix) {
      			b = new byte[len];
      			System.arraycopy(buffer, start, b, 0, len);
      		} 
      		else {
      			b = new byte[len + 4];
      			synchronized(this.bb) {
      				this.bb.putInt(0, len);
      				System.arraycopy(this.bb.array(), 0, b, 0, 4);
      			}
      			System.arraycopy(buffer, start, b, 4, len);
      		}

      		try {
      			return this.queue.offer(b, 100L, TimeUnit.MILLISECONDS);
      		} 
      		catch (InterruptedException ex) {
      			throw new RuntimeException(ex);
      		}
      	}
      	     	
      	public void close() {
      		this.working = false;
      		try {
      			this.out.close();
      		} 
      		catch (IOException ex) {
      			ex.printStackTrace();
      		}
      	}
    }

   
   static class StreamAndSize 
   {
	   long size;
	   InputStream in;
   }
}

